import os
import json
from datetime import timedelta

from django.conf import settings
from django.db import transaction
from django.db.models import Q
from django.http import HttpResponse, JsonResponse
from django.shortcuts import get_object_or_404
from django.utils import timezone
from rest_framework import status, viewsets
from rest_framework.response import Response
from rest_framework.views import APIView

from apps.notebooks.models import Notebook
from apps.notebooks.serializers import NotebookSerializer
from apps.workers.models import (
    Worker,
    Machine,
    WorkerSession,
)
from apps.workers.constants import WorkerState, MachineState, WorkerSessionState
from apps.workers.serializers import WorkerSerializer
from apps.storage.s3utils import clean_worker_files
from apps.accounts.serializers import UserSerializer

MACHINE_SPELL = os.environ.get("MACHINE_SPELL")


class WorkerGetNb(APIView):
    authentication_classes = []

    def get(self, request, session_id, worker_id, notebook_id, format=None):
        try:
            Worker.objects.get(
                pk=worker_id, session_id=session_id, notebook__id=notebook_id
            )
            nb = Notebook.objects.get(pk=notebook_id)
            return Response(NotebookSerializer(nb).data)
        except Exception:
            pass
        return Response(status=status.HTTP_404_NOT_FOUND)


class WorkerGetOwnerAndUser(APIView):
    authentication_classes = []

    def get(self, request, session_id, worker_id, notebook_id, format=None):
        try:
            wrk = Worker.objects.get(
                pk=worker_id, session_id=session_id, notebook__id=notebook_id
            )
            nb = Notebook.objects.get(pk=notebook_id)

            owner = nb.created_by
            user = wrk.run_by
            owner_info = json.loads(owner.profile.info)
            plan = owner_info.get("plan", "starter")

            data = {
                "owner": {
                    "username": owner.username,
                    "email": owner.email,
                    "plan": owner_info.get("plan", "starter"),
                },
                "user": {},
            }
            if user is not None:
                data["user"] = {"username": user.username, "email": user.email}
            return Response(data)
        except Exception:
            pass
        return Response(status=status.HTTP_404_NOT_FOUND)


class WorkerUpdateNb(APIView):
    def post(self, request, session_id, worker_id, notebook_id, format=None):
        try:
            Worker.objects.get(
                pk=worker_id, session_id=session_id, notebook__id=notebook_id
            )
            nb = Notebook.objects.get(pk=notebook_id)
            nb.title = request.data.get("title", nb.title)
            nb.params = request.data.get("params", nb.params)
            nb.save()
            return Response(NotebookSerializer(nb).data)
        except Exception:
            pass
        return Response(status=status.HTTP_404_NOT_FOUND)


class GetWorker(APIView):
    def get(self, request, session_id, worker_id, notebook_id, format=None):
        try:
            worker = Worker.objects.get(
                pk=worker_id, session_id=session_id, notebook__id=notebook_id
            )
            return Response(WorkerSerializer(worker).data)
        except Exception:
            pass
        return Response(status=status.HTTP_404_NOT_FOUND)


class IsWorkerStale(APIView):
    def get(self, request, session_id, worker_id, notebook_id, format=None):
        try:
            worker = Worker.objects.get(
                pk=worker_id, session_id=session_id, notebook__id=notebook_id
            )

            is_stale = worker.updated_at < timezone.now() - timedelta(
                minutes=settings.WORKER_STALE_TIME
            )
            return Response({"is_stale": is_stale})
        except Exception:
            pass
        return Response(status=status.HTTP_404_NOT_FOUND)


class SetWorkerState(APIView):
    def post(self, request, session_id, worker_id, notebook_id, format=None):
        try:
            worker = Worker.objects.get(
                pk=worker_id, session_id=session_id, notebook__id=notebook_id
            )
            worker.state = request.data.get("state", WorkerState.Unknown)
            if request.data.get("machine_id") is not None:
                worker.machine_id = request.data.get("machine_id")
            worker.save()

            sessions = WorkerSession.objects.filter(worker=worker)
            if sessions:
                session = sessions[len(sessions) - 1]
                session.worker = worker
                session.ipv4 = worker.machine_id
                session.state = WorkerSessionState.Running
                session.save()

            return Response(WorkerSerializer(worker).data)
        except Exception:
            pass
        return Response(status=status.HTTP_404_NOT_FOUND)


class DeleteWorker(APIView):
    def post(self, request, session_id, worker_id, notebook_id, format=None):
        try:
            worker = Worker.objects.get(
                pk=worker_id, session_id=session_id, notebook__id=notebook_id
            )
            sessions = WorkerSession.objects.filter(worker=worker)
            if sessions:
                session = sessions[len(sessions) - 1]
                session.worker = None
                session.state = WorkerSessionState.Stopped
                session.save()

            # delete worker output files
            clean_worker_files(worker.notebook.hosted_on.id, worker.session_id)

            worker.delete()

            return Response(status=status.HTTP_204_NO_CONTENT)
        except Exception:
            pass
        return Response(status=status.HTTP_404_NOT_FOUND)


class MachineInfo(APIView):
    def post(self, request, format=None):
        try:
            if MACHINE_SPELL is None:
                return Response(status=status.HTTP_403_FORBIDDEN)
            if MACHINE_SPELL != request.data.get("machine_spell", ""):
                return Response(status=status.HTTP_403_FORBIDDEN)

            ipv4 = request.data.get("ipv4", "")
            if ipv4 == "":
                return Response(status=status.HTTP_400_BAD_REQUEST)

            state = request.data.get("state", "")
            if state not in [s.value for s in MachineState]:
                return Response(status=status.HTTP_400_BAD_REQUEST)

            ms = Machine.objects.filter(ipv4=ipv4)
            if not ms:
                Machine.objects.create(ipv4=ipv4, state=state)
            else:
                for m in ms:
                    m.state = state
                    m.save()

            return Response(status=status.HTTP_204_NO_CONTENT)
        except Exception:
            pass
        return Response(status=status.HTTP_404_NOT_FOUND)



from django.test import TestCase
import requests

from django.contrib.auth.models import User
from apps.accounts.models import Site
from apps.notebooks.models import Notebook
from apps.workers.models import Machine, WorkerSession
from apps.workers.constants import WorkerSessionState
from apps.workers.utils import get_running_machines, shuffle_machines


from datetime import datetime

from allauth.account.admin import EmailAddress
from django.contrib.auth.models import User
from django.core import mail
from django.utils.timezone import make_aware
from rest_framework import status
from rest_framework.authtoken.models import Token
from rest_framework.test import APITestCase

from apps.accounts.models import Membership, Secret, Site
from apps.notebooks.models import Notebook
from apps.workers.models import Worker

# Create your tests here.
# python manage.py test apps.workers -v 2


class ManageMachinesTestCase(TestCase):
    def test_get_running_machines(self):
        ms = get_running_machines()
        self.assertEqual(len(ms), 0)

        Machine.objects.create(ipv4="0.0.0.0", state="Running")
        ms = get_running_machines()
        self.assertEqual(len(ms), 1)

    def test_shuffle_machines(self):
        Machine.objects.create(ipv4="0.0.0.0", state="Running")
        Machine.objects.create(ipv4="0.0.0.0", state="Running")

        ms = get_running_machines()
        self.assertEqual(len(ms), 2)
        ms = shuffle_machines(ms)
        self.assertEqual(len(ms), 2)


class WorkerSessionTestCase(TestCase):
    def setUp(self):
        self.user = User.objects.create_user(
            username="developer",
            email="developer@example.com",
            password="developer",
        )
        EmailAddress.objects.create(
            user=self.user, email=self.user.email, verified=True, primary=True
        )
        self.site = Site.objects.create(
            title="Mercury",
            slug="single-site",
            share=Site.PUBLIC,
            created_by=self.user,
        )

        self.nb = Notebook.objects.create(
            title="some",
            slug="some",
            path="some",
            created_by=self.user,
            hosted_on=self.site,
            file_updated_at=make_aware(datetime.now()),
        )
        self.user2 = User.objects.create_user(
            username="developer2",
            email="developer2@example.com",
            password="developer2",
        )
        EmailAddress.objects.create(
            user=self.user2, email=self.user2.email, verified=True, primary=True
        )

    def test_create(self):
        WorkerSession.objects.create(
            ipv4="0.0.0.0",
            state=WorkerSessionState.Running,
            owned_by=self.user,
            site=self.site,
            notebook=self.nb,
        )
        self.assertEqual(len(WorkerSession.objects.all()), 1)

        WorkerSession.objects.create(
            ipv4="0.0.0.0",
            state=WorkerSessionState.Running,
            owned_by=self.user,
            run_by=self.user2,
            site=self.site,
            notebook=self.nb,
        )
        self.assertEqual(len(WorkerSession.objects.all()), 2)

    def test_update(self):
        WorkerSession.objects.create(
            ipv4="0.0.0.0",
            state=WorkerSessionState.Running,
            owned_by=self.user,
            site=self.site,
            notebook=self.nb,
        )
        self.assertEqual(len(WorkerSession.objects.all()), 1)

        ws = WorkerSession.objects.get(pk=1)
        prev_created_at, prev_updated_at = ws.created_at, ws.updated_at
        ws.save()

        self.assertEqual(prev_created_at, ws.created_at)
        self.assertNotEqual(prev_updated_at, ws.updated_at)


# python manage.py test apps.workers.tests.WorkerGetUserInfoTestCase -v 2
class WorkerGetUserInfoTestCase(TestCase):
    def setUp(self):
        self.user = User.objects.create_user(
            username="developer",
            email="developer@example.com",
            password="developer",
        )
        EmailAddress.objects.create(
            user=self.user, email=self.user.email, verified=True, primary=True
        )
        self.site = Site.objects.create(
            title="Mercury",
            slug="single-site",
            share=Site.PUBLIC,
            created_by=self.user,
        )

        self.nb = Notebook.objects.create(
            title="some",
            slug="some",
            path="some",
            created_by=self.user,
            hosted_on=self.site,
            file_updated_at=make_aware(datetime.now()),
        )

        self.user2 = User.objects.create_user(
            username="developer2",
            email="developer2@example.com",
            password="developer2",
        )
        EmailAddress.objects.create(
            user=self.user2, email=self.user2.email, verified=True, primary=True
        )

    def test_get_owner(self):
        wrk = Worker.objects.create(
            machine_id="some-id", session_id="session-some-id", notebook=self.nb
        )
        url = f"/api/v1/worker/{wrk.session_id}/{wrk.id}/{self.nb.id}/owner-and-user"
        response = self.client.get(url)

        self.assertTrue("owner" in response.data)
        self.assertTrue("username" in response.data["owner"])
        self.assertTrue("email" in response.data["owner"])
        self.assertTrue("plan" in response.data["owner"])
        # there should be user information but it should be empty dict
        self.assertTrue("user" in response.data)
        self.assertTrue(not response.data["user"])  # empty dict = user not logged

    def test_get_user(self):
        wrk = Worker.objects.create(
            machine_id="some-id",
            session_id="session-some-id",
            notebook=self.nb,
            run_by=self.user2,
        )
        url = f"/api/v1/worker/{wrk.session_id}/{wrk.id}/{self.nb.id}/owner-and-user"
        response = self.client.get(url)

        self.assertTrue("user" in response.data)
        self.assertTrue("username" in response.data["user"])
        self.assertTrue("email" in response.data["user"])
